package body A_Stack_M
  with SPARK_Mode,
       --  Although the Stack_Model represents The_Stack as a single
       --  entity, The_Stack is actually refined into two constituents
       Refined_State => (The_Stack => (P, U, V))
is

   -- The types of the objects that are the consistuents of The_Stack
   -- one of which is based on the Stack_Model, but it need not be.
   subtype Stack_Pointer is Stack_Model range 0 .. Stack_Size;
   subtype Index is Stack_Pointer range 1 .. Stack_Size;
   type Vector is array (Index) of Item;

   -- The declaration and initialization of the constituents of The_Stack
   P : Stack_Pointer := 0;
   V : Vector := (others => 0);
   U : Stack_Pointer := 0;

   -- The_Stack function has to translate the refined state into the
   -- Stack_Model.  In this case it is modellled by the number of items in
   -- the stack, which is maintained by the constituent P.
   function The_Stack return Stack_Model is (P) with Refined_Global => P;

   -- Count has to return the number of items in the stack.  Since
   -- The_Stack is modelled by the number of items in the stack, all
   -- that is necessary is to return a Natural obtained from the
   -- Stack_Model parameter.
   function Count (M : Stack_Model) return Natural is (Natural (M));

   -- Element has to return the N'th item on the stack.  Now the refinement
   -- of The_Stack is visible it is possible to index into the Vector, V,
   -- maintaining the items pushed on to The_Stack.
   function Element (M : Stack_Model; N : Positive) return Item is
        (V (Index (N))) with Refined_Global => V;

   function Is_Empty return Boolean is (P = 0) with
      Refined_Global => P;
   function Is_Full return Boolean is (P = Stack_Size) with
      Refined_Global => P;
   function Top return Item is (V (P)) with
      Refined_Global => (P, V);

   -- As the Unchanged proof funtion declared in the package specification
   -- will always return True because of the simplistic Stack_Model.
   -- It has to be shown that the items in The_Stack other than the top item
   -- are unchanged by a call to Push.  A Refined_Post contract can be used
   -- to specify this requirement.
   -- A Refined_Post contract is specified in terms of the constituents of
   -- The_Stack, P and V.  As V has type Vector which is an array type the
   -- SPARK defined attribute, 'Update can be used to specify this invariance.
   -- A'Update (Index => Value) returns an array identical to A except that
   -- component A (Index) = Value.
   procedure Push (It : in Item) with
     Refined_Global => (In_Out => (P, U, V)),
     -- We nedd to show V (1 .. P'Old) is invariant
     -- 'Update provides this mechanism for arrays
     Refined_Post => V = V'Old'Update (P => It) and
                     P = P'Old + 1
   is
   begin
      P := P + 1;
      V (P) := It;
      if P > U then
         U := P;
      end if;
   end Push;

   -- It is also necessary to demonstrate that following a call to Pop
   -- the items in The_Stack other than the top item are unchange.
   -- In the refinement of The_Stack, V is a Vector maintaining the items on
   -- The_Stack. V has a Refined_Global mode of Input this ensures that the
   -- elements of V are invariant over a call to Pop and hence no
   -- Refined_Post contract is needed.
   procedure Pop (It : out Item) with
     Refined_Global => (In_Out => P,
                        Input  => (V, U))
   is
   begin
      It := V (P);
      P := P - 1;
   end Pop;

   function Utilization return Natural is (Natural (U)) with
     Refined_Global => U;

end A_Stack_M;
