package body Dictionary
  with SPARK_Mode,
       Refined_State => (State => (Dict_Count, Dict_Entries))
is

   subtype Key_Length_Type is Natural range 0 .. Max_Key_Length;
   subtype Key_Range is Natural range 1 .. Max_Key_Length;
   subtype Key_String is String (Key_Range);
   subtype Count_Range is Natural range 0 .. Capacity;
   subtype Index is Count_Range range 1 .. Count_Range'Last;

   type Key_Type is record
      Length : Key_Length_Type;
      The_Key : Key_String;
   end record;

   type Dictionary_Entry is record
      Key : Key_Type;
      El  : Element;
   end record;

   type Entry_Array is array (Index) of Dictionary_Entry;

   Dict_Count : Count_Range;
   Dict_Entries : Entry_Array;

   function Count return Natural is (Dict_Count) with
   Refined_Global => Dict_Count;

   function Equal_Keys (K1, K2 : Key_Type) return Boolean is
     (K1.Length = K2.Length and then
      (for all I in 1 .. K1.Length => K1.The_Key (I) = K2.The_Key (I)));

   function To_Key_Type (S : String) return Key_Type
   is
      Result : Key_Type := (Length => 0, The_Key => (others => ' '));
   begin
      for I in S'Range loop
         exit when Result.Length = Result.The_Key'Last;
         Result.Length := Result.Length + 1;
         Result.The_Key (Result.Length) := S (I);
      end loop;
      return Result;
   end To_Key_Type;

   function Has (X : Element) return Boolean is
     (for some I in 1 .. Dict_Count => Dict_Entries (I).El = X) with
   Refined_Global => (Dict_Count, Dict_Entries);

   -- The implementation of Item_Key is a existential quantification.
   -- If the Key exists it must be related to the given Item to return True.
   function Item_Key (Item : Element; Key : String) return Boolean is
     (for some I in 1 .. Dict_Count =>
         Equal_Keys (Dict_Entries (I).Key,
                     To_Key_Type (Key)) and then
         Dict_Entries (I).El = Item) with
     Refined_Global => (Dict_Count, Dict_Entries);

   function Item (Key : String) return Element with
     Refined_Global => (Dict_Count, Dict_Entries)
   is
      Local_Key : constant Key_Type := To_Key_Type (Key);
      Next_Entry : Dictionary_Entry;
      It : Element := Null_Element;
   begin
      for I in 1 .. Dict_Count loop
         Next_Entry := Dict_Entries (I);
         if Equal_Keys (Next_Entry.Key, Local_Key) then
            It := Next_Entry.El;
            exit;
         end if;
         pragma Loop_Invariant (It = Null_Element);
      end loop;
      return It;
   end Item;

   procedure Put (X : Element; Key : String) with
     Refined_Global => (In_Out => (Dict_Count, Dict_Entries))
   is
   begin
      -- Do not add to the dictionary X is already present.
      if not Has (X) then
         Dict_Count := Dict_Count + 1;
         Dict_Entries (Dict_Count) :=
           (Key => To_Key_Type (Key),
            El => X);
         -- Assertion helps with the proof of the postcondition
         pragma Assert (Dict_Entries (Dict_Count).Key = To_Key_Type (Key));
      end if;
   end Put;

   procedure Dummy_Init_Entries (Dict_Entries : out Entry_Array);

   procedure Dummy_Init_Entries (Dict_Entries : out Entry_Array) with
     SPARK_Mode => Off
   is
   begin
      null;
   end Dummy_Init_Entries;
begin
   Dict_Count := 0;
   Dummy_Init_Entries (Dict_Entries);
end Dictionary;
