package Dictionary
  with SPARK_Mode,
       Abstract_State    => State,
       Initializes       => State,
       Initial_Condition => Count = 0
is

   Capacity : constant Natural := 2000;
   Max_Key_Length : constant := 32;

   type Element is private;
   Null_Element : constant Element;

   function Count return Natural with
     Global => State,
     Ghost;

   -- For GNATprove it is preferable that functions that appear in
   -- pre and postconditions are implemented as expression functions.
   -- Also functions which return a Boolean result are easier to
   -- write if the expression is quantified.
   -- The function Item is difficult to write as an expression function so
   -- this Boolean proof function has been provided and is True if the
   -- the dictionary contains the given Key associated with the given Item.
   -- Look at the postcondition of the function Item.
   function Item_Key (Item : Element; Key : String) return Boolean with
     Global => State,
     Ghost;

   function Has (X : Element) return Boolean with
     Global => State;

   -- The original example did not specify what happens if the Key does not
   -- exist in the dictionary.
   -- This function returns a Null_Element if the Key is not present.
   function Item (Key : String) return Element with
     Global => State,
     Post => Item_Key (Item'Result, Key) or else
             Item'Result = Null_Element;

   procedure Put (X : Element; Key : String) with
     Global => (In_Out => State),
     Pre => Count < Capacity,
     Contract_Cases => (not Has (X) =>
                            Has (X) and
                            Item_Key (X, Key) and
                            Count = Count'Old + 1,
                        Has (X) =>
                            Has (X) = Has (X)'Old and
                            Item_Key (X, Key) = Item_Key (X, Key)'Old and
                            Count = Count'Old);
private
   -- This is just an example of what an element might be.
   -- It could be any non-limited SPARK type.
   -- The package could be generic with this type as a formal parameter
   type Element is record
      X, Y : Natural;
   end record;
   -- The warnings off should not be required.
   -- It is a shortcoming with the GPL tools.
   pragma Warnings (Off);
   Null_Element : constant Element := (0, 0);
   pragma Warnings (On);
   -- Make sure warnings are turned on again.
end Dictionary;
