with Ada.Text_IO;   use Ada.Text_IO;
with Display;       use Display;
with Display.Basic; use Display.Basic;

procedure Main is

   --  Logical coordinate system
   X_Size : constant Natural := 8;
   Y_Size : constant Natural := 16;  
   subtype X_Coord is Natural range 1 .. X_Size;
   subtype Y_Coord is Natural range 1 .. Y_Size;
   
   type Cell_Type is (Empty, I, O, J, L, S, T, Z);
   -- I: cyan, J: blue, L: white, O: yellow, S: green, T: magenta, Z: red
   
   subtype Piece_Type is Cell_Type range I .. Z;
   
   type Cell is record
      Kind     : Cell_Type;
      Graphics : Shape_Id;
   end record;
   
   type Piece is record
      Kind     : Piece_Type;
      Y        : Y_Coord;
      X        : X_Coord;
      Graphics : Shape_Id;
   end record;
   
   type Row   is array (X_Coord) of Cell;
   type Board is array (Y_Coord) of Row;
   
   --  Pixel constants for graphical components
   Pxl_Width    : constant Natural := 10;
   Pxl_Origin_X : constant Integer := -40;
   Pxl_Origin_Y : constant Integer := 80;
   
   Game_Board : Board;
   
   Falling_Piece : Piece;
   
   function New_Board return Board 
   is
   begin
      return Board' (others => (others => Cell' (Kind     => Empty, 
                                                 Graphics => Null_Shape_Id)));
   end New_Board;
      
   function New_Piece (Kind : Piece_Type) return Piece 
   is
      Color   : Color_Type;
      Start_Y : constant Y_Coord := 1;
      Start_X : constant X_Coord := X_Size / 2;
   begin
      case Kind is
         when I => Color := Cyan;
         when J => Color := Blue;
         when L => Color := White;
         when O => Color := Yellow;
         when S => Color := Green;
         when T => Color := Magenta;
         when Z => Color := Red;
      end case;
      return (Kind     => Kind,
              X        => Start_X,
              Y        => Start_Y,
              Graphics => 
                New_Box (X      => Float (Pxl_Origin_X + Start_X * Pxl_Width),
                         Y      => Float (Pxl_Origin_Y - Start_Y * Pxl_Width),
                         Width  => Float (Pxl_Width),
                         Height => Float (Pxl_Width),
                         Color  => Color));
   end New_Piece;
   
   procedure Insert_Piece_In_Board (P: in Piece; B: in out Board) is 
   begin
      B (P.Y) (P.X) := (Kind     => P.Kind, 
                        Graphics => P.Graphics);
   end Insert_Piece_In_Board;
   
   procedure Update_Graphics (P: in out Piece) is
   begin
      Set_X (P.Graphics, Float (Pxl_Origin_X + P.X * Pxl_Width));
      Set_Y (P.Graphics, Float (Pxl_Origin_Y - P.Y * Pxl_Width));
   end Update_Graphics;
   
   procedure Print_Board (B : in Board) is 
   begin
      for Y in Y_Coord loop
         for X in X_Coord loop
            Put (if B (Y) (X).Kind = Empty then "- " else 
                    Cell_Type'Image (B (Y) (X).Kind) & " ");   
         end loop;
         Put_Line (" ");
      end loop;
      Put_Line (" ");
   end Print_Board;
   
begin
   
   Game_Board := New_Board;
   
   Falling_Piece := New_Piece (S);
   
   for Y in Y_Coord loop
      delay 0.2;
      Falling_Piece.Y := Y;
      Update_Graphics (Falling_Piece);
   end loop;
   
   Insert_Piece_In_Board (P => Falling_Piece,
                          B => Game_Board);
   
   Print_Board (Game_Board);
   Put_Line ("GAME OVER!");

end Main;
