with Ada.Text_IO;   use Ada.Text_IO;
package body Tetris
with SPARK_Mode => On
is
   --  Pixel constants for graphical components
   Pxl_Side     : constant Natural := 10;
   Pxl_Origin_X : constant Integer := -40;
   Pxl_Origin_Y : constant Integer := 80;

   function New_Board return Board
   is
   begin
      return Board' (others => (others => Cell' (Kind    => Empty,
                                                 Graphic => Null_Shape_Id)));
   end New_Board;

   function New_Piece (Kind : Piece_Type) return Piece
   is
      Color   : Color_Type;
      Start_Y : constant Y_Coord := 1;
      Start_X : constant X_Coord := X_Size / 2;
   begin
      Color := (case Kind is
                   when I => Cyan,
                   when J => Blue,
                   when L => White,
                   when O => Yellow,
                   when S => Green,
                   when T => Magenta,
                   when Z => Red);

      return Piece' (Kind     => Kind,
                     X        => Start_X,
                     Y        => Start_Y,
                     Graphic =>
                       New_Box (X => Float (Pxl_Origin_X + Start_X * Pxl_Side),
                                Y => Float (Pxl_Origin_Y - Start_Y * Pxl_Side),
                                Width  => Float (Pxl_Side),
                                Height => Float (Pxl_Side),
                                Color  => Color));
   end New_Piece;

   procedure Insert_Piece_In_Board (Cur_Piece: in Piece;
                                    Cur_Board: in out Board) is
   begin
      Cur_Board (Cur_Piece.Y) (Cur_Piece.X) := (Kind    => Cur_Piece.Kind,
                                                Graphic => Cur_Piece.Graphic);
   end Insert_Piece_In_Board;

   procedure Update_Graphics (P: in out Piece) is
   begin
      Set_X (P.Graphic, Float (Pxl_Origin_X + P.X * Pxl_Side));
      Set_Y (P.Graphic, Float (Pxl_Origin_Y - P.Y * Pxl_Side));
   end Update_Graphics;

   procedure Do_Action (A       : in     Action;
                        P       : in out Piece;
                        B       : in     Board;
                        Success : out Boolean) is
      Candidate : Piece;
   begin
      if Move_Is_Within_Board (P, A) then
         Candidate := Move (P, A);
         if No_Overlap (B, Candidate) then
            P := Candidate;
            Success := True;
            return;
         end if;
      end if;
      Success := False;
   end Do_Action;

   procedure Delete_Complete_Line (Cur_Board : in out Board) is
      From_Line : Y_Coord;
   begin
      if Is_Complete_Line (Cur_Board (Y_Size)) then
         for Col in X_Coord loop
            if Cur_Board (Y_Size) (Col).Kind /= Empty then
                  Delete (Shape => Cur_Board (Y_Size) (Col).Graphic);
            end if;
         end loop;
       for To_Line in reverse 2 .. Y_Size loop
            From_Line := To_Line -1;
            for Col in X_Coord loop
               Cur_Board (To_Line) (Col) := Cur_Board (From_Line) (Col);
               if Cur_Board (To_Line) (Col).Kind /= Empty then
                  Set_Y (Shape => Cur_Board (To_Line) (Col). Graphic,
                         Value => Float (Pxl_Origin_Y - To_Line * Pxl_Side));
               end if;
            end loop;
         end loop;
         for Col in X_Coord loop
            Cur_Board (Y_Coord'First) (Col).Kind := Empty;
            pragma Loop_Invariant (for all C in X_Coord'First .. Col =>
                                     Cur_Board (Y_Coord'First) (C).Kind = Empty);
         end loop;
      end if;
   end Delete_Complete_Line;

   procedure Row_Has_Piece_Type (PT    : in Piece_Type;
                                 Row   : in Y_Coord;
                                 B     : in Board;
                                 Found : out Boolean;
                                 Col   : out X_Coord)
   is
      X : Natural range X_Coord'First .. X_Coord'Last + 1;
   begin
      Found := False;
      Col   := X_Coord'First;
      X := X_Coord'First;
      while not Found and X <= X_Coord'Last loop
         pragma Loop_Invariant ((if Found then B (Row) (X).Kind = PT)
                                and (for all C in X_Coord'First .. X - 1 =>
                                       B (Row) (C).Kind /= PT));
         if B (Row) (X).Kind = PT then
            Found := True;
            Col   := X;
         end if;
         X := X + 1;
      end loop;

   end Row_Has_Piece_Type;

end Tetris;
