with Display;       use Display;
with Display.Basic; use Display.Basic;

package Tetris
with SPARK_Mode => On
is

   --  Logical coordinate system for the board:
   --  Y_Size rows, X_Size cells, origin (1, 1) at top left corner
   X_Size : constant Natural := 8;
   Y_Size : constant Natural := 16;
   subtype X_Coord is Natural range 1 .. X_Size;
   subtype Y_Coord is Natural range 1 .. Y_Size;

   --  Cell contents is either Empty or a Tetris piece
   --  (I: cyan, J: blue, L: white, O: yellow, S: green, T: magenta, Z: red)
   type Cell_Type is (Empty, I, O, J, L, S, T, Z);

   --  Tetris piece type
   subtype Piece_Type is Cell_Type range I .. Z;

   type Cell is record
      Kind    : Cell_Type;
      Graphic : Shape_Id;
   end record;

   type Piece is record
      Kind    : Piece_Type;
      Y       : Y_Coord;
      X       : X_Coord;
      Graphic : Shape_Id;
   end record;

   type Row   is array (X_Coord) of Cell;
   type Board is array (Y_Coord) of Row;

   type Action is (Move_Left, Move_Right, Move_None, Move_Down);
   subtype User_Action is Action range Move_Left .. Move_None;

   function Is_Empty (B : Board; Y : Integer; X : Integer) return Boolean is
      (X in X_Coord and then Y in Y_Coord and then B(Y)(X). Kind = Empty);

   function Move_Is_Within_Board (P : Piece; A : Action) return Boolean is
      (case A is
          when Move_Left   => P.X - 1 in X_Coord,
          when Move_Right  => P.X + 1 in X_Coord,
          when Move_None   => True,
          when Move_Down   => P.Y + 1 in Y_Coord);

   function Move (P : Piece; A : Action) return Piece is
     (case A is
         when Move_Left   => P'Update (X => P.X - 1),
         when Move_Right  => P'Update (X => P.X + 1),
         when Move_None  =>  P,
         when Move_Down   => P'Update (Y => P.Y + 1))
   with Pre => Move_Is_Within_Board (P, A);

   function No_Overlap (B : Board; P : Piece) return Boolean is
      (Is_Empty (B, P.Y, P.X));

   procedure Do_Action (A       : in     Action;
                        P       : in out Piece;
                        B       : in     Board;
                        Success : out Boolean);

   function New_Board return Board;

   function New_Piece (Kind : Piece_Type) return Piece;

   procedure Insert_Piece_In_Board (Cur_Piece: in Piece;
                                    Cur_Board: in out Board);

   procedure Update_Graphics (P: in out Piece);

   function Game_Over (B : Board) return Boolean is
     (B (Y_Coord'First) (X_Size / 2). Kind /= Empty);

   function Is_Complete_Line (L : Row) return Boolean is
     (for all X in X_Coord => L(X). Kind /= Empty);

   function Is_Empty_Line (L : Row) return Boolean is
     (for all X in X_Coord => L(X) . Kind = Empty);

   function Equal_Lines (L1, L2 : Row) return Boolean is
     (for all X in X_Coord => L1(X) . Kind = L2 (X). Kind);

   procedure Delete_Complete_Line (Cur_Board : in out Board)
     with Post =>
       (Is_Complete_Line (Cur_Board'Old (Y_Size)) and
          Is_Empty_Line(Cur_Board (Y_Coord'First)))
     or Cur_Board = Cur_Board'Old ;

   procedure Row_Has_Piece_Type (PT    : in Piece_Type;
                                 Row   : in Y_Coord;
                                 B     : in Board;
                                 Found : out Boolean;
                                 Col   : out X_Coord)
     with Post =>
       (Found = (for some C in X_Coord =>
                       B (Row) (C).Kind = PT))
       and
       (if Found then B (Row) (Col).Kind = PT
        and (for all C in X_Coord'First .. Col -1 =>
                   B (Row) (C).Kind /= PT));
end Tetris;
