package A_Stack_M
with SPARK_Mode,
  Abstract_State => The_Stack,
  Initializes => The_Stack,
-- The initial condition specifies status of the state absttraction
-- immediately after package elabotation.
-- Here it shows the program and logical status.
  Initial_Condition => Is_Empty and Count (The_Stack) = 0
is
   Stack_Size : constant := 100;

   -- A model type is declared for The_Stack.
   -- This type should not be available for general use.
   -- The type is indefinite and objects of this type
   -- (where the full type declaration is not visible)
   -- cannot declared without inialization
   type Stack_Model (<>) is private;
   subtype Item is Integer range 0 .. 20;

   -- This is the only constant or function that has a result type of
   -- Stack_Model.  It is a ghost function and cannot be used to initialize
   -- object declarations.  This means that objects of type Stack_Model
   -- cannot be declared outside of this package or its children.
   -- It is an abstract logical proof function representing the value of
   -- The_Stack.
   function The_Stack return Stack_Model with
     Global => The_Stack,
     Convention => Ghost;

   -- Abstract, ghost proof function used in pre and postconditions and other
   -- assertion expressions representing the number of items on the stack.
   -- It has a Stack_Model parameter.
   function Count (M : Stack_Model) return Natural with
     Convention => Ghost;

   -- Abstract, ghost proof function allowing iteration over the itmes on
   -- a stack.  It has a Stack_Model parameter and a Positive parameter for
   -- selecting the Nth element of the stack.
   -- M is used in the preconditon of Element.
   pragma Warnings (Off, "unused initial value of ""M""");
   function Element (M : Stack_Model; N : Positive) return Item with
     Global => The_Stack,
     Pre => N <= Count (M),
     Convention => Ghost;
   pragma Warnings (On, "unused initial value of ""M""");

   -- Unchanged is a abstract, ghost proof function which will always be True
   -- provided its precondition is satisfied because of simplifications in the
   -- Stack_Model.
   -- Unchanged is used to describe invariance of the elements of the stack for
   -- use in proving the funtionality of clients which call the operations on
   -- The_Stack.
   -- Because the function will always return True, it is necessary to prove
   -- this invariance holds in the implementation of Push and Pop.
   -- Unchanged has two Stack_Parameter parameters, the Invariant_Part
   -- representing the unchanged part of the stack and New_Stack representing
   -- the updated stack which shauld include the Invariant_Part.
   -- The function is written in terms of the proof functions Count and Element.
   function Unchanged (Invariant_Part,
                       New_Stack : Stack_Model) return Boolean is
     (for all I in 1 .. Count (Invariant_Part) =>
           Element (Invariant_Part, I) = Element (New_Stack, I)) with
     Global => The_Stack,
     Pre => Count (Invariant_Part) <= Count (New_Stack),
     Convention => Ghost;

   -- Note the program functions relate the program and logical abstractions
   -- in their pre and postconditions.
   function Is_Empty return Boolean with
     Global => The_Stack,
     Post => Is_Empty'Result = (Count (The_Stack) = 0);
   function Is_Full return Boolean with
     Global => The_Stack,
   Post => Is_Full'Result = (Count (The_Stack) = Stack_Size);
   function Top return Item with
     Global => The_Stack,
     Pre  => not Is_Empty or else Count (The_Stack) > 0,
     Post => Top'Result = Element (The_Stack, Count (The_Stack));

   -- The pre and postconditions of push and pop describe their use of both
   -- the logical and program abstractions
   procedure Push (It : in Item) with
     Global => (In_Out => The_Stack),
     Pre => not Is_Full or else Count (The_Stack) < Stack_Size,
     Post => Top = It and
             Unchanged (The_Stack'Old, The_Stack) and
             Count (The_Stack) = Count (The_Stack)'Old + 1 and
             Element (The_Stack, Count (The_Stack)) = It;

   procedure Pop (It : out Item) with
     Global => (In_Out => The_Stack),
     Pre => not Is_Empty or else Count (The_Stack) > 0,
     Post => It = Top'Old and
             Unchanged (The_Stack, The_Stack'Old) and
             Count (The_Stack) = Count (The_Stack)'Old - 1 and
             It = Element (The_Stack, Count (The_Stack))'Old;

private
   -- A simplified model is used to represent The_Stack, just the number
   -- of items in the stack.  This avoids copying large data structures but
   -- it does mean that extra checks have to be performed in the implementation
   -- to ensure that this simplification is justified.
   type Stack_Model is range 0 .. Stack_Size;
end A_Stack_M;
